package events

import (
	"fmt"
	"runtime/debug"
	"sync"

	"github.com/libp2p/go-eventbus"
	"github.com/libp2p/go-libp2p-core/event"
	"go.uber.org/zap/zapcore"

	"github.com/spacemeshos/go-spacemesh/common/types"
	"github.com/spacemeshos/go-spacemesh/log"
	"github.com/spacemeshos/go-spacemesh/timesync"
)

// Subscription is a subscription to events.
// Consumer must be aware that publish will block if subsription is not read fast enough.
type Subscription = event.Subscription

var (
	mu sync.RWMutex
	// reporter is the event reporter singleton.
	reporter *EventReporter
)

// InitializeReporter initializes the event reporting interface with
// a nonzero channel buffer. This is useful for testing, where we want reporting to
// block.
func InitializeReporter() {
	mu.Lock()
	defer mu.Unlock()
	if reporter != nil {
		return
	}
	reporter = newEventReporter()
}

// EventHook returns hook for logger.
func EventHook() func(entry zapcore.Entry) error {
	return func(entry zapcore.Entry) error {
		// If we report anything less than this we'll end up in an infinite loop
		if entry.Level >= zapcore.ErrorLevel {
			ReportError(NodeError{
				Msg:   entry.Message,
				Trace: string(debug.Stack()),
				Level: entry.Level,
			})
		}
		return nil
	}
}

// ReportNewTx dispatches incoming events to the reporter singleton.
func ReportNewTx(layerID types.LayerID, tx *types.Transaction) {
	ReportTxWithValidity(layerID, tx, true)
}

// ReportTxWithValidity reports a tx along with whether it was just invalidated.
func ReportTxWithValidity(layerID types.LayerID, tx *types.Transaction, valid bool) {
	mu.RLock()
	defer mu.RUnlock()
	txWithValidity := Transaction{
		Transaction: tx,
		LayerID:     layerID,
		Valid:       valid,
	}
	if reporter != nil {
		if err := reporter.transactionEmitter.Emit(txWithValidity); err != nil {
			// TODO(nkryuchkov): consider returning an error and log outside the function
			log.With().Error("Failed to emit transaction", tx.ID(), layerID, log.Err(err))
		} else {
			log.Debug("reported tx: %v", txWithValidity)
		}
	}
}

// ReportNewActivation reports a new activation.
func ReportNewActivation(activation *types.ActivationTx) {
	mu.RLock()
	defer mu.RUnlock()

	activationTxEvent := ActivationTx{ActivationTx: activation}

	if reporter != nil {
		innerBytes, err := activation.InnerBytes()
		if err != nil {
			log.Error("error attempting to report activation: unable to encode activation")
			return
		}

		if err := reporter.activationEmitter.Emit(activationTxEvent); err != nil {
			// TODO(nkryuchkov): consider returning an error and log outside the function
			log.With().Error("Failed to emit activation", activation.ID(), activation.PubLayerID, log.Err(err))
		} else {
			log.With().Debug("reported activation", activation.Fields(len(innerBytes))...)
		}
	}
}

// ReportRewardReceived reports a new reward.
func ReportRewardReceived(r Reward) {
	mu.RLock()
	defer mu.RUnlock()

	if reporter != nil {
		if err := reporter.rewardEmitter.Emit(r); err != nil {
			// TODO(nkryuchkov): consider returning an error and log outside the function
			log.With().Error("Failed to emit rewards", r.Layer, log.Err(err))
		} else {
			log.Debug("reported reward: %v", r)
		}
	}
}

// ReportLayerUpdate reports a new layer, or an update to an existing layer.
func ReportLayerUpdate(layer LayerUpdate) {
	mu.RLock()
	defer mu.RUnlock()

	if reporter != nil {
		if err := reporter.layerEmitter.Emit(layer); err != nil {
			// TODO(nkryuchkov): consider returning an error and log outside the function
			log.With().Error("Failed to emit updated layer", layer, log.Err(err))
		} else {
			log.With().Debug("reported new or updated layer", layer)
		}
	}
}

// ReportError reports an error.
func ReportError(err NodeError) {
	mu.RLock()
	defer mu.RUnlock()

	if reporter != nil {
		if err := reporter.errorEmitter.Emit(err); err != nil {
			// TODO(nkryuchkov): consider returning an error and log outside the function
			log.With().Error("Failed to emit error", log.Err(err))
		} else {
			log.Debug("reported error: %v", err)
		}
	}
}

// ReportNodeStatusUpdate reports an update to the node status. It just
// pings the listener to notify them that there is an update; the listener
// is responsible for fetching the new status details. This is because
// status contains disparate information coming from different services,
// and the listener already knows how to gather that information so there
// is no point in duplicating that logic here.
// Note: There is some overlap with channelLayer here, as a new latest
// or verified layer should be sent over that channel as well. However,
// that happens inside the Mesh, at the source. It doesn't currently
// happen here because the status update includes only a layer ID, not
// full layer data, and the Reporter currently has no way to retrieve
// full layer data.
func ReportNodeStatusUpdate() {
	mu.RLock()
	defer mu.RUnlock()

	if reporter != nil {
		if err := reporter.statusEmitter.Emit(Status{}); err != nil {
			// TODO(nkryuchkov): consider returning an error and log outside the function
			log.With().Error("Failed to emit status update", log.Err(err))
		} else {
			log.Debug("reported status update")
		}
	}
}

// ReportReceipt reports creation or receipt of a new tx receipt.
func ReportReceipt(r TxReceipt) {
	mu.RLock()
	defer mu.RUnlock()

	if reporter != nil {
		if err := reporter.receiptEmitter.Emit(r); err != nil {
			// TODO(nkryuchkov): consider returning an error and log outside the function
			log.With().Error("Failed to emit receipt", r.ID, log.Err(err))
		} else {
			log.Debug("reported receipt: %v", r)
		}
	}
}

// ReportAccountUpdate reports an account whose data has been updated.
func ReportAccountUpdate(a types.Address) {
	mu.RLock()
	defer mu.RUnlock()

	accountEvent := Account{Address: a}

	if reporter != nil {
		if err := reporter.accountEmitter.Emit(accountEvent); err != nil {
			// TODO(nkryuchkov): consider returning an error and log outside the function
			log.With().Error("Failed to emit account update", log.String("account", a.Short()), log.Err(err))
		} else {
			log.With().Debug("reported account update", a)
		}
	}
}

// SubscribeTxs subscribes to new transactions.
func SubscribeTxs() Subscription {
	mu.RLock()
	defer mu.RUnlock()

	if reporter != nil {
		sub, err := reporter.bus.Subscribe(new(Transaction))
		if err != nil {
			log.With().Panic("Failed to subscribe to transactions")
		}

		return sub
	}
	return nil
}

// SubscribeActivations subscribes to activations.
func SubscribeActivations() Subscription {
	mu.RLock()
	defer mu.RUnlock()

	if reporter != nil {
		sub, err := reporter.bus.Subscribe(new(ActivationTx))
		if err != nil {
			log.With().Panic("Failed to subscribe to activations")
		}

		return sub
	}
	return nil
}

// SubscribeLayers subscribes to all layer data.
func SubscribeLayers() Subscription {
	mu.RLock()
	defer mu.RUnlock()

	if reporter != nil {
		sub, err := reporter.bus.Subscribe(new(LayerUpdate))
		if err != nil {
			log.With().Panic("Failed to subscribe to layers")
		}

		return sub
	}
	return nil
}

// SubscribeErrors subscribes to node errors.
func SubscribeErrors() Subscription {
	mu.RLock()
	defer mu.RUnlock()

	if reporter != nil {
		sub, err := reporter.bus.Subscribe(new(NodeError))
		if err != nil {
			log.With().Panic("Failed to subscribe to errors")
		}

		return sub
	}
	return nil
}

// SubscribeStatus subscribes to node status messages.
func SubscribeStatus() Subscription {
	mu.RLock()
	defer mu.RUnlock()

	if reporter != nil {
		sub, err := reporter.bus.Subscribe(new(Status))
		if err != nil {
			log.With().Panic("Failed to subscribe to status")
		}

		return sub
	}
	return nil
}

// SubscribeAccount subscribes to account data updates.
func SubscribeAccount() Subscription {
	mu.RLock()
	defer mu.RUnlock()

	if reporter != nil {
		sub, err := reporter.bus.Subscribe(new(Account))
		if err != nil {
			log.With().Panic("Failed to subscribe to account updates")
		}

		return sub
	}
	return nil
}

// SubscribeRewards subscribes to rewards.
func SubscribeRewards() Subscription {
	mu.RLock()
	defer mu.RUnlock()

	if reporter != nil {
		sub, err := reporter.bus.Subscribe(new(Reward))
		if err != nil {
			log.With().Panic("Failed to subscribe to rewards")
		}

		return sub
	}
	return nil
}

// SubscribeReceipts subscribes to receipts.
func SubscribeReceipts() Subscription {
	mu.RLock()
	defer mu.RUnlock()

	if reporter != nil {
		sub, err := reporter.bus.Subscribe(new(TxReceipt))
		if err != nil {
			log.With().Panic("Failed to subscribe to receipts")
		}

		return sub
	}
	return nil
}

// SubscribeToLayers is used to track and report automatically every time a
// new layer is reached.
func SubscribeToLayers(newLayerCh timesync.LayerTimer) {
	mu.RLock()
	defer mu.RUnlock()

	if reporter != nil {
		// This will block, so run in a goroutine
		go func() {
			for {
				mu.RLock()
				stopChan := reporter.stopChan
				mu.RUnlock()

				select {
				case layer := <-newLayerCh:
					log.With().Debug("reporter got new layer", layer)
					ReportNodeStatusUpdate()
				case <-stopChan:
					return
				}
			}
		}()
	}
}

// The status of a layer
// TODO: this list is woefully inadequate and does not map to reality. See https://github.com/spacemeshos/api/issues/144.
const (
	LayerStatusTypeUnknown   = iota
	LayerStatusTypeApproved  // approved by Hare
	LayerStatusTypeConfirmed // confirmed by Tortoise
)

// LayerUpdate packages up a layer with its status (which a layer does not ordinarily contain).
type LayerUpdate struct {
	LayerID types.LayerID
	Status  int
}

// Field returns a log field. Implements the LoggableField interface.
func (nl LayerUpdate) Field() log.Field {
	return log.String("layer", fmt.Sprintf("status: %d, number: %d", nl.Status, nl.LayerID))
}

// NodeError represents an internal error to be reported.
type NodeError struct {
	Msg   string
	Trace string
	Level zapcore.Level
}

// TxReceipt represents a transaction receipt.
type TxReceipt struct {
	ID      types.TransactionID
	Result  int
	GasUsed uint64
	Fee     uint64
	Layer   types.LayerID
	Index   uint32
	Address types.Address
}

// Reward represents a reward object with extra data needed by the API.
type Reward struct {
	Layer       types.LayerID
	Total       uint64
	LayerReward uint64
	Coinbase    types.Address
	// TODO: We don't currently have a way to get the Layer Computed.
	// See https://github.com/spacemeshos/go-spacemesh/issues/2275
	// LayerComputed
}

// Transaction wraps a tx with its layer ID and validity info.
type Transaction struct {
	Transaction *types.Transaction
	LayerID     types.LayerID
	Valid       bool
}

// ActivationTx wraps *types.ActivationTx.
type ActivationTx struct {
	*types.ActivationTx
}

// Status indicates status change event.
type Status struct{}

// Account wraps account address.
type Account struct {
	types.Address
}

// EventReporter is the struct that receives incoming events and dispatches them.
type EventReporter struct {
	bus                event.Bus
	transactionEmitter event.Emitter
	activationEmitter  event.Emitter
	layerEmitter       event.Emitter
	errorEmitter       event.Emitter
	statusEmitter      event.Emitter
	accountEmitter     event.Emitter
	rewardEmitter      event.Emitter
	receiptEmitter     event.Emitter
	proposalsEmitter   event.Emitter
	stopChan           chan struct{}
}

func newEventReporter() *EventReporter {
	bus := eventbus.NewBus()

	transactionEmitter, err := bus.Emitter(new(Transaction))
	if err != nil {
		log.With().Panic("failed to create transaction emitter", log.Err(err))
	}

	activationEmitter, err := bus.Emitter(new(ActivationTx))
	if err != nil {
		log.With().Panic("failed to create activation emitter", log.Err(err))
	}

	layerEmitter, err := bus.Emitter(new(LayerUpdate))
	if err != nil {
		log.With().Panic("failed to create layer emitter", log.Err(err))
	}

	statusEmitter, err := bus.Emitter(new(Status))
	if err != nil {
		log.With().Panic("failed to create status emitter", log.Err(err))
	}

	accountEmitter, err := bus.Emitter(new(Account))
	if err != nil {
		log.With().Panic("failed to create account emitter", log.Err(err))
	}

	rewardEmitter, err := bus.Emitter(new(Reward))
	if err != nil {
		log.With().Panic("failed to create reward emitter", log.Err(err))
	}

	receiptEmitter, err := bus.Emitter(new(TxReceipt))
	if err != nil {
		log.With().Panic("failed to create receipt emitter", log.Err(err))
	}

	errorEmitter, err := bus.Emitter(new(NodeError))
	if err != nil {
		log.With().Panic("failed to create error emitter", log.Err(err))
	}

	proposalsEmitter, err := bus.Emitter(new(EventProposal))
	if err != nil {
		log.With().Panic("failed to to create proposal emitter", log.Err(err))
	}

	return &EventReporter{
		bus:                bus,
		transactionEmitter: transactionEmitter,
		activationEmitter:  activationEmitter,
		layerEmitter:       layerEmitter,
		statusEmitter:      statusEmitter,
		accountEmitter:     accountEmitter,
		rewardEmitter:      rewardEmitter,
		receiptEmitter:     receiptEmitter,
		errorEmitter:       errorEmitter,
		proposalsEmitter:   proposalsEmitter,
		stopChan:           make(chan struct{}),
	}
}

// CloseEventReporter shuts down the event reporting service and closes open channels.
func CloseEventReporter() {
	mu.Lock()
	defer mu.Unlock()
	if reporter != nil {
		if err := reporter.transactionEmitter.Close(); err != nil {
			log.With().Panic("failed to close transactionEmitter", log.Err(err))
		}
		if err := reporter.activationEmitter.Close(); err != nil {
			log.With().Panic("failed to close activationEmitter", log.Err(err))
		}
		if err := reporter.layerEmitter.Close(); err != nil {
			log.With().Panic("failed to close layerEmitter", log.Err(err))
		}
		if err := reporter.errorEmitter.Close(); err != nil {
			log.With().Panic("failed to close errorEmitter", log.Err(err))
		}
		if err := reporter.statusEmitter.Close(); err != nil {
			log.With().Panic("failed to close statusEmitter", log.Err(err))
		}
		if err := reporter.accountEmitter.Close(); err != nil {
			log.With().Panic("failed to close accountEmitter", log.Err(err))
		}
		if err := reporter.rewardEmitter.Close(); err != nil {
			log.With().Panic("failed to close rewardEmitter", log.Err(err))
		}
		if err := reporter.receiptEmitter.Close(); err != nil {
			log.With().Panic("failed to close receiptEmitter", log.Err(err))
		}
		if err := reporter.proposalsEmitter.Close(); err != nil {
			log.With().Panic("failed to close propoposalsEmitter", log.Err(err))
		}

		close(reporter.stopChan)
		reporter = nil
	}
}
